/*
	SOS Emulator Add-on template
	(c) Mikhail Panasyuk aka BeAR, 2009. E-mail: otoko@yandex.ru

	Contains portions of NewOS kernel code:

	Copyright 2001-2002, Travis Geiselbrecht. All rights reserved. 
	Distributed under the terms of the NewOS License. 
*/

#include "../../sos_emulator_addon.h"

#include <stdlib.h>

const char* add_on_name = "BeOS2";

// The run queue. Holds the threads ready to run ordered by priority.
static struct thread_queue gRunQueue = {NULL, NULL};

// The current thread being run.
static struct thread *_currThread = NULL;

// fake functions
struct thread* thread_get_current_thread(void)
{
	if (!_currThread)
		_currThread = gRunQueue.head;

	return _currThread;
}
void context_switch(struct thread *oldThread, struct thread *nextThread)
{
	_currThread = nextThread;
}

static int _rand(void)
{
   static int next = 0;
   if (next == 0) next = system_time();
   next = next * 1103515245 + 12345;
   return((next >> 16) & 0x7FFF);
}

void scheduler_enqueue_in_run_queue(struct thread *thread)
{
	struct thread *curr, *prev;

	for(curr = gRunQueue.head, prev = NULL; 
		curr && (curr->priority >= thread->priority); 
		curr = curr->queue_next) 
	{
		if (prev)
			prev = prev->queue_next;
		else
			prev = gRunQueue.head;
	}

	thread->queue_next = curr;

	if(prev)
		prev->queue_next = thread;
	else
		gRunQueue.head = thread;

}

void scheduler_remove_from_run_queue(struct thread *thread)
{
   struct thread *item, *prev;

   // find thread in run queue
   for (item = gRunQueue.head, prev = NULL; 
        item && item != thread; 
        item = item->queue_next) 
   {
      if (prev)
         prev = prev->queue_next;
      else
         prev = gRunQueue.head;
   }
   
   if (prev)
      prev->queue_next = item->queue_next;
   else
      gRunQueue.head = item->queue_next;
}

struct thread* select_next_thread_to_run()
{
	struct thread *nextThread, *nextLower;

	nextThread = gRunQueue.head;
	
     // always extract real time threads
	if (nextThread->priority >= B_FIRST_REAL_TIME_PRIORITY) {
		return nextThread;
	}
	
	if (nextThread->priority == B_IDLE_PRIORITY) {
		return nextThread;
	}
	while (1) {
		// find thread with next lower priority
		nextLower = nextThread->queue_next;
		while (nextLower->priority >= nextThread->priority)
			nextLower = nextLower->queue_next;

     	// never skip last non-idle normal subqueue of threads
     	if (nextLower->priority == B_IDLE_PRIORITY)
			break;
		
		// skip subqueue of normal threads sometimes
		if(rand()%2) break;

		nextThread = nextLower;
	}
	return nextThread;
}

void scheduler_reschedule(void)
{
  struct thread *oldThread = thread_get_current_thread();
  struct thread *nextThread;

  switch (oldThread->next_state) 
  {
     case B_THREAD_RUNNING:
     case B_THREAD_READY:
        scheduler_enqueue_in_run_queue(oldThread);
        break;

/*     case B_THREAD_SUSPENDED:
        break;

     case THREAD_STATE_FREE_ON_RESCHED:
        // This will hopefully be eliminated once the slab
        // allocator is done
        thread_enqueue(oldThread, &dead_q);
        break;
*/
     default:
        break;
  }

  oldThread->state = oldThread->next_state;

  // select next thread from the run queue
  nextThread = select_next_thread_to_run();

  // extract selected thread from the run queue
  scheduler_remove_from_run_queue(nextThread);
  
  nextThread->state = B_THREAD_RUNNING;
  nextThread->next_state = B_THREAD_READY;

  if (nextThread != oldThread/* || oldThread->cpu->info.preempted*/) 
  {

/*     bigtime_t quantum = 3000; // ToDo: calculate quantum!
     timer *quantum_timer= &oldThread->cpu->info.quantum_timer;

     if (!oldThread->cpu->info.preempted)
        _local_timer_cancel_event(oldThread->cpu->info.cpu_num, quantum_timer);
     
     oldThread->cpu->info.preempted = 0;
     add_timer(quantum_timer, &reschedule_event, quantum, B_ONE_SHOT_RELATIVE_TIMER);
*/
     if (nextThread != oldThread)
        context_switch(oldThread, nextThread);
  }
}
